"""Converts a list of ECP tokens into a python AST and provides the environment for execution."""
from parsergen import *
from parsergen.parser import ParseError
from .lexer import *
from .tracker import Tracer
import sys, os
try:
    import astor
except ImportError:
    print("astor module not found - will not be able to convert ECP to python source code")
    astor = None
from math import sqrt
from random import randint
from ast import *
import ast
from ecp.parser import EcpParser
from parsergen.parser_utils import *
_List = ast.List
_Dict = ast.Dict

BUILTIN_IMPORT = "from ecp.topython import *\n"

def _dump(node, annotate_fields=True, include_attributes=False, *, indent=None):
    """
    Return a formatted dump of the tree in node.  This is mainly useful for
    debugging purposes.  If annotate_fields is true (by default),
    the returned string will show the names and the values for fields.
    If annotate_fields is false, the result string will be more compact by
    omitting unambiguous field names.  Attributes such as line
    numbers and column offsets are not dumped by default.  If this is wanted,
    include_attributes can be set to true.  If indent is a non-negative
    integer or string, then the tree will be pretty-printed with that indent
    level. None (the default) selects the single line representation.
    """
    def _format(node, level=0):
        if indent is not None:
            level += 1
            prefix = '\n' + indent * level
            sep = ',\n' + indent * level
        else:
            prefix = ''
            sep = ', '
        if isinstance(node, AST):
            cls = type(node)
            args = []
            allsimple = True
            keywords = annotate_fields
            for name in node._fields:
                try:
                    value = getattr(node, name)
                except AttributeError:
                    keywords = True
                    continue
                if value is None and getattr(cls, name, ...) is None:
                    keywords = True
                    continue
                value, simple = _format(value, level)
                allsimple = allsimple and simple
                if keywords:
                    args.append('%s=%s' % (name, value))
                else:
                    args.append(value)
            if include_attributes and node._attributes:
                for name in node._attributes:
                    try:
                        value = getattr(node, name)
                    except AttributeError:
                        continue
                    if value is None and getattr(cls, name, ...) is None:
                        continue
                    value, simple = _format(value, level)
                    allsimple = allsimple and simple
                    args.append('%s=%s' % (name, value))
            if allsimple and len(args) <= 3:
                return '%s(%s)' % (node.__class__.__name__, ', '.join(args)), not args
            return '%s(%s%s)' % (node.__class__.__name__, prefix, sep.join(args)), False
        elif isinstance(node, list):
            if not node:
                return '[]', True
            return '[%s%s]' % (prefix, sep.join(_format(x, level)[0] for x in node)), False
        return repr(node), True

    if not isinstance(node, AST):
        raise TypeError('expected AST, got %r' % node.__class__.__name__)
    if indent is not None and not isinstance(indent, str):
        indent = ' ' * indent
    return _format(node)[0]


class Namespace:
    def __init__(self, **kwargs) -> None:
        for k, v in kwargs.items():
            setattr(self, k, v)


def fix_line_and_column(node):
    for child in walk(node):
        if 'lineno' in child._attributes:
            child.lineno = getattr(child, 'lineno', getattr(node, 'lineno', 0))
        if 'end_lineno' in child._attributes:
            child.end_lineno = getattr(child, 'end_lineno') or 0
            if child.end_lineno < child.lineno:
                child.end_lineno = getattr(child, "lineno", 0)
        
        if 'col_offset' in child._attributes:
            child.col_offset = getattr(child, 'col_offset', getattr(node, 'col_offset', 0))
        if 'end_col_offset' in child._attributes:
            child.end_col_offset = getattr(child, 'end_col_offset') or 0
            if child.end_col_offset < child.col_offset:
                child.end_col_offset = getattr(child, "col_offset", 0)
    
    return fix_missing_locations(node)

def parse_ecp(text: str, mode="exec"):
    lex_result = EcpLexer().lex_string(text)
    tokens = TokenStream(lex_result)
    p = EcpParser(tokens)
    rv = p.program()
    error = p.error()
    if rv is None and error is not None:
        raise error
    rv = fix_line_and_column(rv)
    if mode == "single":
        rv = Interactive(body=rv.body)
    return rv

def get_more(text: str) -> bool:
    code = None
    
    try:
        code = compile(parse_ecp(text, mode="single"), "<interactive>", "single")
    except ParseError as e:
        pass
    
    if code or text.endswith("\n"):
        return False
    else:
        return True
    

def to_py_source(text: Union[str, Module]):
    code = text
    if isinstance(text, str):
        code = parse_ecp(text)
    if astor is not None:
        return BUILTIN_IMPORT + astor.to_source(code)
    raise Exception("astor module not found - cannot convert ecp to python source code")

def ecp(text: str=None, *, file: str=None, name="<unkown>", showAST=False, scope=None, trace=None, tracecompact=False, mode="exec"):
    if text is None:
        with open(file, encoding="utf-8") as f:
            text = f.read()
    if scope is None:
        scope = {}
    if isinstance(scope, Namespace):
        scope = vars(scope)
    if trace is None:
        trace = []
    scope.update(globals())
    r = parse_ecp(text, mode=mode)
    if showAST:
        print(_dump(r, include_attributes=True, indent=2))
    if len(trace) > 0:
        with Tracer(trace, compact=tracecompact):
            exec(compile(parse(r, mode=mode), name, mode), scope)
    else:
        exec(compile(parse(r, mode=mode), name, mode), scope)

    return Namespace(**scope)

def _ECP_IMPORT(location: str, target: str, scope=None):
    m = None
    for p in sys.path:
        if os.path.exists(p + location + ".ecp"):
            s = ecp(file=p + location + ".ecp", scope=globals())
            scope[target] = s
            return
    raise ImportError(name=location, path=p + location + ".ecp")


# ECP BUILTINS

_MAGIC_OUTPUT = print
_MAGIC_USERINPUT = input
INPUT = input
LEN = len
Integer = int
Int = int
Real = float
Bool = bool
String = str
Array = list
Dictionary = dict

def POSITION(string: str, to_match: str) -> int:
    try:
        return string.index(to_match)
    except:
        return -1

def SUBSTRING(start: int, end: int, string: str):
    return string[start:end+1]

STRING_TO_INT = int
STRING_TO_REAL = float
INT_TO_STRING = REAL_TO_STRING = str
CHAR_TO_CODE = ord
CODE_TO_CHAR = chr
RANDOM_INT = randint
SQRT = sqrt

PY = exec


from .parser_helpers import *

# Code @generated by parsergen; do not edit!
from parsergen.parser_utils import GeneratedParser, TokenStream, Node, Filler
from parsergen.parser_utils import memoize, memoize_left_rec
from functools import reduce
class EcpParser(GeneratedParser):
    @memoize
    def program(self):
        pos = self.mark()
        """
        p=compound EOF { Module(body=p, type_ignores=[], **self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.compound()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('EOF')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            p = parts[0]
            return Module(body=p, type_ignores=[], **self.loc)
        self.goto(pos)
        
        return None
        
    @memoize
    def compound(self):
        pos = self.mark()
        """
        c=statement* { PyECP_Compound(c, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self._loop_0()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            c = parts[0]
            return PyECP_Compound(c, self.loc)
        self.goto(pos)
        
        return None
        
    def _loop_0(self):
        """
        statement*
        """
        children = []
        while True:
            pos = self.mark()
            part = self.statement()
            if self.match(part): children.append(part)
            else:
                self.goto(pos)
                break
        return children
    @memoize
    def statement(self):
        pos = self.mark()
        """
        e=if_statement | for_loop | while_loop | repeat_until_loop | record_definition | try_catch | suboroutine_definition | class_definition | import_statement | assignment_statement { e };
        """
        parts = []
        for _ in range(1):
            part = self._or_1()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            e = parts[0]
            return e
        self.goto(pos)
        
        """
        e=expr { PyECP_ExprStatement(e) };
        """
        parts = []
        for _ in range(1):
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            e = parts[0]
            return PyECP_ExprStatement(e)
        self.goto(pos)
        
        return None
        
    def _or_1(self):
        """
        if_statement | for_loop | while_loop | repeat_until_loop | record_definition | try_catch | suboroutine_definition | class_definition | import_statement | assignment_statement
        """
        pos = self.mark()
        part = self.if_statement()
        if self.match(part): return part
        self.goto(pos)
        part = self.for_loop()
        if self.match(part): return part
        self.goto(pos)
        part = self.while_loop()
        if self.match(part): return part
        self.goto(pos)
        part = self.repeat_until_loop()
        if self.match(part): return part
        self.goto(pos)
        part = self.record_definition()
        if self.match(part): return part
        self.goto(pos)
        part = self.try_catch()
        if self.match(part): return part
        self.goto(pos)
        part = self.suboroutine_definition()
        if self.match(part): return part
        self.goto(pos)
        part = self.class_definition()
        if self.match(part): return part
        self.goto(pos)
        part = self.import_statement()
        if self.match(part): return part
        self.goto(pos)
        part = self.assignment_statement()
        if self.match(part): return part
        self.goto(pos)
        self.fail()
        return None
    @memoize
    def assignment_statement(self):
        pos = self.mark()
        """
        target=variable (COLON ID)? ASSIGN value=expr { PyECP_Assign(target, value, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.variable()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._maybe_2()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('ASSIGN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            target = parts[0]
            value = parts[3]
            return PyECP_Assign(target, value, self.loc)
        self.goto(pos)
        
        return None
        
    def _maybe_2(self):
        """
        (COLON ID)?
        """
        pos = self.mark()
        part = self._expr_list_3()
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    def _expr_list_3(self):
        """
        (COLON ID)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expect('COLON')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('ID')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    @memoize
    def variable(self):
        pos = self.mark()
        """
        CONSTANT? name=ID indexing=indexing { PyECP_Variable(name, indexing, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self._maybe_4()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('ID')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.indexing()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            name = parts[1]
            indexing = parts[2]
            return PyECP_Variable(name, indexing, self.loc)
        self.goto(pos)
        
        return None
        
    def _maybe_4(self):
        """
        CONSTANT?
        """
        pos = self.mark()
        part = self.expect('CONSTANT')
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    @memoize
    def parameters(self):
        pos = self.mark()
        """
        params=(expr !ASSIGN (COMMA expr !ASSIGN)* COMMA?)? { PyECP_Parameters(params) };
        """
        parts = []
        for _ in range(1):
            part = self._maybe_5()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            params = parts[0]
            return PyECP_Parameters(params)
        self.goto(pos)
        
        return None
        
    def _maybe_5(self):
        """
        (expr !ASSIGN (COMMA expr !ASSIGN)* COMMA?)?
        """
        pos = self.mark()
        part = self._expr_list_6()
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    def _expr_list_6(self):
        """
        (expr !ASSIGN (COMMA expr !ASSIGN)* COMMA?)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            predicate_pos = self.mark()
            part = self.expect('ASSIGN')
            self.goto(predicate_pos)
            if self.match(part):
                self.fail()
                break
            part = self._loop_7()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._maybe_8()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    def _loop_7(self):
        """
        (COMMA expr !ASSIGN)*
        """
        children = []
        while True:
            pos = self.mark()
            part = self._expr_list_9()
            if self.match(part): children.append(part)
            else:
                self.goto(pos)
                break
        return children
    def _expr_list_9(self):
        """
        (COMMA expr !ASSIGN)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expect('COMMA')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            predicate_pos = self.mark()
            part = self.expect('ASSIGN')
            self.goto(predicate_pos)
            if self.match(part):
                self.fail()
                break
            return parts
        self.goto(pos)
        return None
    def _maybe_8(self):
        """
        COMMA?
        """
        pos = self.mark()
        part = self.expect('COMMA')
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    @memoize
    def kw_parameters(self):
        pos = self.mark()
        """
        params=(ID ASSIGN expr (COMMA ID ASSIGN expr)* COMMA?)? { PyECP_KwParameters(params) };
        """
        parts = []
        for _ in range(1):
            part = self._maybe_10()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            params = parts[0]
            return PyECP_KwParameters(params)
        self.goto(pos)
        
        return None
        
    def _maybe_10(self):
        """
        (ID ASSIGN expr (COMMA ID ASSIGN expr)* COMMA?)?
        """
        pos = self.mark()
        part = self._expr_list_11()
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    def _expr_list_11(self):
        """
        (ID ASSIGN expr (COMMA ID ASSIGN expr)* COMMA?)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expect('ID')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('ASSIGN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._loop_12()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._maybe_13()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    def _loop_12(self):
        """
        (COMMA ID ASSIGN expr)*
        """
        children = []
        while True:
            pos = self.mark()
            part = self._expr_list_14()
            if self.match(part): children.append(part)
            else:
                self.goto(pos)
                break
        return children
    def _expr_list_14(self):
        """
        (COMMA ID ASSIGN expr)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expect('COMMA')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('ID')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('ASSIGN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    def _maybe_13(self):
        """
        COMMA?
        """
        pos = self.mark()
        part = self.expect('COMMA')
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    @memoize
    def attr_index(self):
        pos = self.mark()
        """
        DOT i=ID { "attr", i.value };
        """
        parts = []
        for _ in range(1):
            part = self.expect('DOT')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('ID')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            i = parts[1]
            return "attr", i.value
        self.goto(pos)
        
        return None
        
    @memoize
    def subscript_index(self):
        pos = self.mark()
        """
        LS_PAREN i=expr RS_PAREN { "subscript", i };
        """
        parts = []
        for _ in range(1):
            part = self.expect('LS_PAREN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('RS_PAREN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            i = parts[1]
            return "subscript", i
        self.goto(pos)
        
        return None
        
    @memoize
    def call(self):
        pos = self.mark()
        """
        LPAREN params=parameters kw_params=kw_parameters RPAREN { "call", (params, kw_params) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('LPAREN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.parameters()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.kw_parameters()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('RPAREN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            params = parts[1]
            kw_params = parts[2]
            return "call", (params, kw_params)
        self.goto(pos)
        
        return None
        
    @memoize
    def indexing(self):
        pos = self.mark()
        """
        indexes=(attr_index | subscript_index | call)* { [i for [i] in indexes] };
        """
        parts = []
        for _ in range(1):
            part = self._loop_15()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            indexes = parts[0]
            return [i for [i] in indexes]
        self.goto(pos)
        
        return None
        
    def _loop_15(self):
        """
        (attr_index | subscript_index | call)*
        """
        children = []
        while True:
            pos = self.mark()
            part = self._expr_list_16()
            if self.match(part): children.append(part)
            else:
                self.goto(pos)
                break
        return children
    def _expr_list_16(self):
        """
        (attr_index | subscript_index | call)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self._or_17()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    def _or_17(self):
        """
        attr_index | subscript_index | call
        """
        pos = self.mark()
        part = self.attr_index()
        if self.match(part): return part
        self.goto(pos)
        part = self.subscript_index()
        if self.match(part): return part
        self.goto(pos)
        part = self.call()
        if self.match(part): return part
        self.goto(pos)
        self.fail()
        return None
    @memoize_left_rec
    def factor(self):
        pos = self.mark()
        """
        f=factor_part i=indexing { PyECP_Factor(f, i, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.factor_part()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.indexing()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            f = parts[0]
            i = parts[1]
            return PyECP_Factor(f, i, self.loc)
        self.goto(pos)
        
        return None
        
    @memoize_left_rec
    def factor_part(self):
        pos = self.mark()
        """
        t=INT | FLOAT | BOOLEAN | STRING | NONE { PyECP_Constant(t, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self._or_18()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            t = parts[0]
            return PyECP_Constant(t, self.loc)
        self.goto(pos)
        
        """
        LPAREN e=expr RPAREN { e };
        """
        parts = []
        for _ in range(1):
            part = self.expect('LPAREN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('RPAREN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            e = parts[1]
            return e
        self.goto(pos)
        
        """
        p=array | dictionary | tuple | magic_function | variable { p };
        """
        parts = []
        for _ in range(1):
            part = self._or_19()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            p = parts[0]
            return p
        self.goto(pos)
        
        """
        op=PLUS | SUB | NOT e=factor { PyECP_UnaryOp(op, e) };
        """
        parts = []
        for _ in range(1):
            part = self._or_20()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.factor()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            op = parts[0]
            e = parts[1]
            return PyECP_UnaryOp(op, e)
        self.goto(pos)
        
        return None
        
    def _or_18(self):
        """
        INT | FLOAT | BOOLEAN | STRING | NONE
        """
        pos = self.mark()
        part = self.expect('INT')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('FLOAT')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('BOOLEAN')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('STRING')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('NONE')
        if self.match(part): return part
        self.goto(pos)
        self.fail()
        return None
    def _or_19(self):
        """
        array | dictionary | tuple | magic_function | variable
        """
        pos = self.mark()
        part = self.array()
        if self.match(part): return part
        self.goto(pos)
        part = self.dictionary()
        if self.match(part): return part
        self.goto(pos)
        part = self.tuple()
        if self.match(part): return part
        self.goto(pos)
        part = self.magic_function()
        if self.match(part): return part
        self.goto(pos)
        part = self.variable()
        if self.match(part): return part
        self.goto(pos)
        self.fail()
        return None
    def _or_20(self):
        """
        PLUS | SUB | NOT
        """
        pos = self.mark()
        part = self.expect('PLUS')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('SUB')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('NOT')
        if self.match(part): return part
        self.goto(pos)
        self.fail()
        return None
    @memoize_left_rec
    def expr(self):
        pos = self.mark()
        """
        e=op_or { e };
        """
        parts = []
        for _ in range(1):
            part = self.op_or()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            e = parts[0]
            return e
        self.goto(pos)
        
        return None
        
    @memoize_left_rec
    def op_or(self):
        pos = self.mark()
        """
        base=op_and others=(OR op_and)+ { PyECP_BoolOp(base, "OR", others, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.op_and()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._loop_21()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            base = parts[0]
            others = parts[1]
            return PyECP_BoolOp(base, "OR", others, self.loc)
        self.goto(pos)
        
        """
        e=op_and { e };
        """
        parts = []
        for _ in range(1):
            part = self.op_and()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            e = parts[0]
            return e
        self.goto(pos)
        
        return None
        
    def _loop_21(self):
        """
        (OR op_and)+
        """
        children = []
        while True:
            pos = self.mark()
            part = self._expr_list_22()
            if self.match(part): children.append(part)
            else:
                if len(children) == 0:
                    self.fail()
                self.goto(pos)
                break
        return children if len(children) > 0 else None
    def _expr_list_22(self):
        """
        (OR op_and)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expect('OR')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.op_and()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    @memoize_left_rec
    def op_and(self):
        pos = self.mark()
        """
        base=inversion others=(AND inversion)+ { PyECP_BoolOp(base, "AND", others, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.inversion()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._loop_23()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            base = parts[0]
            others = parts[1]
            return PyECP_BoolOp(base, "AND", others, self.loc)
        self.goto(pos)
        
        """
        e=inversion { e };
        """
        parts = []
        for _ in range(1):
            part = self.inversion()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            e = parts[0]
            return e
        self.goto(pos)
        
        return None
        
    def _loop_23(self):
        """
        (AND inversion)+
        """
        children = []
        while True:
            pos = self.mark()
            part = self._expr_list_24()
            if self.match(part): children.append(part)
            else:
                if len(children) == 0:
                    self.fail()
                self.goto(pos)
                break
        return children if len(children) > 0 else None
    def _expr_list_24(self):
        """
        (AND inversion)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expect('AND')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.inversion()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    @memoize_left_rec
    def inversion(self):
        pos = self.mark()
        """
        op=NOT e=inversion { PyECP_UnaryOp(op, e) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('NOT')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.inversion()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            op = parts[0]
            e = parts[1]
            return PyECP_UnaryOp(op, e)
        self.goto(pos)
        
        """
        e=comparison { e };
        """
        parts = []
        for _ in range(1):
            part = self.comparison()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            e = parts[0]
            return e
        self.goto(pos)
        
        return None
        
    @memoize_left_rec
    def comparison(self):
        pos = self.mark()
        """
        base=sum others=(LT | GT | LE | GE | EQ | NE sum)+ { PyECP_Comparison(base, others, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.sum()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._loop_25()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            base = parts[0]
            others = parts[1]
            return PyECP_Comparison(base, others, self.loc)
        self.goto(pos)
        
        """
        e=sum { e };
        """
        parts = []
        for _ in range(1):
            part = self.sum()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            e = parts[0]
            return e
        self.goto(pos)
        
        return None
        
    def _loop_25(self):
        """
        (LT | GT | LE | GE | EQ | NE sum)+
        """
        children = []
        while True:
            pos = self.mark()
            part = self._expr_list_26()
            if self.match(part): children.append(part)
            else:
                if len(children) == 0:
                    self.fail()
                self.goto(pos)
                break
        return children if len(children) > 0 else None
    def _expr_list_26(self):
        """
        (LT | GT | LE | GE | EQ | NE sum)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self._or_27()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.sum()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    def _or_27(self):
        """
        LT | GT | LE | GE | EQ | NE
        """
        pos = self.mark()
        part = self.expect('LT')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('GT')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('LE')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('GE')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('EQ')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('NE')
        if self.match(part): return part
        self.goto(pos)
        self.fail()
        return None
    @memoize_left_rec
    def sum(self):
        pos = self.mark()
        """
        left=sum op=ADD | SUB right=term { PyECP_BinOp(left, right, op, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.sum()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._or_28()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.term()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            left = parts[0]
            op = parts[1]
            right = parts[2]
            return PyECP_BinOp(left, right, op, self.loc)
        self.goto(pos)
        
        """
        e=term { e };
        """
        parts = []
        for _ in range(1):
            part = self.term()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            e = parts[0]
            return e
        self.goto(pos)
        
        return None
        
    def _or_28(self):
        """
        ADD | SUB
        """
        pos = self.mark()
        part = self.expect('ADD')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('SUB')
        if self.match(part): return part
        self.goto(pos)
        self.fail()
        return None
    @memoize_left_rec
    def term(self):
        pos = self.mark()
        """
        left=term op=MUL | DIV | INT_DIV | MOD right=uop { PyECP_BinOp(left, right, op, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.term()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._or_29()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.uop()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            left = parts[0]
            op = parts[1]
            right = parts[2]
            return PyECP_BinOp(left, right, op, self.loc)
        self.goto(pos)
        
        """
        e=uop { e };
        """
        parts = []
        for _ in range(1):
            part = self.uop()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            e = parts[0]
            return e
        self.goto(pos)
        
        return None
        
    def _or_29(self):
        """
        MUL | DIV | INT_DIV | MOD
        """
        pos = self.mark()
        part = self.expect('MUL')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('DIV')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('INT_DIV')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('MOD')
        if self.match(part): return part
        self.goto(pos)
        self.fail()
        return None
    @memoize_left_rec
    def uop(self):
        pos = self.mark()
        """
        op=ADD | SUB e=uop { PyECP_UnaryOp(op, e) };
        """
        parts = []
        for _ in range(1):
            part = self._or_30()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.uop()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            op = parts[0]
            e = parts[1]
            return PyECP_UnaryOp(op, e)
        self.goto(pos)
        
        """
        e=power { e };
        """
        parts = []
        for _ in range(1):
            part = self.power()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            e = parts[0]
            return e
        self.goto(pos)
        
        return None
        
    def _or_30(self):
        """
        ADD | SUB
        """
        pos = self.mark()
        part = self.expect('ADD')
        if self.match(part): return part
        self.goto(pos)
        part = self.expect('SUB')
        if self.match(part): return part
        self.goto(pos)
        self.fail()
        return None
    @memoize_left_rec
    def power(self):
        pos = self.mark()
        """
        left=power op=POW right=factor { PyECP_BinOp(left, right, op, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.power()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('POW')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.factor()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            left = parts[0]
            op = parts[1]
            right = parts[2]
            return PyECP_BinOp(left, right, op, self.loc)
        self.goto(pos)
        
        """
        e=factor { e };
        """
        parts = []
        for _ in range(1):
            part = self.factor()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            e = parts[0]
            return e
        self.goto(pos)
        
        return None
        
    @memoize
    def magic_function(self):
        pos = self.mark()
        """
        name=MAGIC parameters=parameters { PyECP_Magic(name.value, parameters, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('MAGIC')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.parameters()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            name = parts[0]
            parameters = parts[1]
            return PyECP_Magic(name.value, parameters, self.loc)
        self.goto(pos)
        
        return None
        
    @memoize
    def param_definition(self):
        pos = self.mark()
        """
        name=ID (COLON ID)? { arg(arg=name.value, annotation=None, **self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('ID')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._maybe_31()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            name = parts[0]
            return arg(arg=name.value, annotation=None, **self.loc)
        self.goto(pos)
        
        return None
        
    def _maybe_31(self):
        """
        (COLON ID)?
        """
        pos = self.mark()
        part = self._expr_list_32()
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    def _expr_list_32(self):
        """
        (COLON ID)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expect('COLON')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('ID')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    @memoize
    def suboroutine_definition(self):
        pos = self.mark()
        """
        SUBROUTINE name=ID LPAREN params=(param_definition (COMMA param_definition)* COMMA?)? RPAREN block=compound END { PyECP_SubroutineDef(name, params, block, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('SUBROUTINE')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('ID')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('LPAREN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._maybe_33()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('RPAREN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.compound()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('END')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            name = parts[1]
            params = parts[3]
            block = parts[5]
            return PyECP_SubroutineDef(name, params, block, self.loc)
        self.goto(pos)
        
        return None
        
    def _maybe_33(self):
        """
        (param_definition (COMMA param_definition)* COMMA?)?
        """
        pos = self.mark()
        part = self._expr_list_34()
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    def _expr_list_34(self):
        """
        (param_definition (COMMA param_definition)* COMMA?)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.param_definition()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._loop_35()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._maybe_36()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    def _loop_35(self):
        """
        (COMMA param_definition)*
        """
        children = []
        while True:
            pos = self.mark()
            part = self._expr_list_37()
            if self.match(part): children.append(part)
            else:
                self.goto(pos)
                break
        return children
    def _expr_list_37(self):
        """
        (COMMA param_definition)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expect('COMMA')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.param_definition()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    def _maybe_36(self):
        """
        COMMA?
        """
        pos = self.mark()
        part = self.expect('COMMA')
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    @memoize
    def if_statement(self):
        pos = self.mark()
        """
        i=_if_statement END { i };
        """
        parts = []
        for _ in range(1):
            part = self._if_statement()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('END')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            i = parts[0]
            return i
        self.goto(pos)
        
        return None
        
    @memoize
    def _if_statement(self):
        pos = self.mark()
        """
        IF condition=expr THEN block=compound other=(elseif_statement | else_statement)? { PyECP_IfStatement(condition, block, other, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('IF')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('THEN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.compound()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._maybe_38()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            condition = parts[1]
            block = parts[3]
            other = parts[4]
            return PyECP_IfStatement(condition, block, other, self.loc)
        self.goto(pos)
        
        return None
        
    def _maybe_38(self):
        """
        (elseif_statement | else_statement)?
        """
        pos = self.mark()
        part = self._expr_list_39()
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    def _expr_list_39(self):
        """
        (elseif_statement | else_statement)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self._or_40()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    def _or_40(self):
        """
        elseif_statement | else_statement
        """
        pos = self.mark()
        part = self.elseif_statement()
        if self.match(part): return part
        self.goto(pos)
        part = self.else_statement()
        if self.match(part): return part
        self.goto(pos)
        self.fail()
        return None
    @memoize
    def elseif_statement(self):
        pos = self.mark()
        """
        ELSE otherwise=_if_statement { [otherwise] };
        """
        parts = []
        for _ in range(1):
            part = self.expect('ELSE')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._if_statement()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            otherwise = parts[1]
            return [otherwise]
        self.goto(pos)
        
        return None
        
    @memoize
    def else_statement(self):
        pos = self.mark()
        """
        ELSE block=compound { block };
        """
        parts = []
        for _ in range(1):
            part = self.expect('ELSE')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.compound()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            block = parts[1]
            return block
        self.goto(pos)
        
        return None
        
    @memoize
    def while_loop(self):
        pos = self.mark()
        """
        WHILE condition=expr block=compound END { PyECP_While(condition, block, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('WHILE')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.compound()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('END')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            condition = parts[1]
            block = parts[2]
            return PyECP_While(condition, block, self.loc)
        self.goto(pos)
        
        return None
        
    @memoize
    def repeat_until_loop(self):
        pos = self.mark()
        """
        REPEAT block=compound UNTIL condition=expr { PyECP_RepeatUntil(condition, block, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('REPEAT')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.compound()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('UNTIL')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            block = parts[1]
            condition = parts[3]
            return PyECP_RepeatUntil(condition, block, self.loc)
        self.goto(pos)
        
        return None
        
    @memoize
    def array(self):
        pos = self.mark()
        """
        LS_PAREN values=parameters RS_PAREN { PyECP_Array(values, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('LS_PAREN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.parameters()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('RS_PAREN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            values = parts[1]
            return PyECP_Array(values, self.loc)
        self.goto(pos)
        
        return None
        
    @memoize
    def tuple(self):
        pos = self.mark()
        """
        LPAREN values=parameters RPAREN { PyECP_Tuple(values, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('LPAREN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.parameters()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('RPAREN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            values = parts[1]
            return PyECP_Tuple(values, self.loc)
        self.goto(pos)
        
        return None
        
    @memoize
    def dictionary(self):
        pos = self.mark()
        """
        LC_BRACE kv_pairs=(expr COLON expr (COMMA expr COLON expr)* COMMA?)? RC_BRACE { PyECP_Dictionary(kv_pairs, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('LC_BRACE')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._maybe_41()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('RC_BRACE')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            kv_pairs = parts[1]
            return PyECP_Dictionary(kv_pairs, self.loc)
        self.goto(pos)
        
        return None
        
    def _maybe_41(self):
        """
        (expr COLON expr (COMMA expr COLON expr)* COMMA?)?
        """
        pos = self.mark()
        part = self._expr_list_42()
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    def _expr_list_42(self):
        """
        (expr COLON expr (COMMA expr COLON expr)* COMMA?)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('COLON')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._loop_43()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._maybe_44()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    def _loop_43(self):
        """
        (COMMA expr COLON expr)*
        """
        children = []
        while True:
            pos = self.mark()
            part = self._expr_list_45()
            if self.match(part): children.append(part)
            else:
                self.goto(pos)
                break
        return children
    def _expr_list_45(self):
        """
        (COMMA expr COLON expr)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expect('COMMA')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('COLON')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    def _maybe_44(self):
        """
        COMMA?
        """
        pos = self.mark()
        part = self.expect('COMMA')
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    @memoize
    def for_loop(self):
        pos = self.mark()
        """
        FOR v=variable ASSIGN start=expr TO end=expr step=(STEP expr)? block=compound END { PyECP_ForTo(v, start, end, step, block, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('FOR')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.variable()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('ASSIGN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('TO')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._maybe_46()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.compound()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('END')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            v = parts[1]
            start = parts[3]
            end = parts[5]
            step = parts[6]
            block = parts[7]
            return PyECP_ForTo(v, start, end, step, block, self.loc)
        self.goto(pos)
        
        """
        FOR v=variable IN iterator=expr block=compound END { PyECP_ForIn(v, iterator, block, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('FOR')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.variable()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('IN')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.compound()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('END')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            v = parts[1]
            iterator = parts[3]
            block = parts[4]
            return PyECP_ForIn(v, iterator, block, self.loc)
        self.goto(pos)
        
        return None
        
    def _maybe_46(self):
        """
        (STEP expr)?
        """
        pos = self.mark()
        part = self._expr_list_47()
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    def _expr_list_47(self):
        """
        (STEP expr)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expect('STEP')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    @memoize
    def record_definition(self):
        pos = self.mark()
        """
        RECORD name=ID values=(variable (COLON ID)?)* END { PyECP_Record(name, values, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('RECORD')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('ID')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._loop_48()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('END')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            name = parts[1]
            values = parts[2]
            return PyECP_Record(name, values, self.loc)
        self.goto(pos)
        
        return None
        
    def _loop_48(self):
        """
        (variable (COLON ID)?)*
        """
        children = []
        while True:
            pos = self.mark()
            part = self._expr_list_49()
            if self.match(part): children.append(part)
            else:
                self.goto(pos)
                break
        return children
    def _expr_list_49(self):
        """
        (variable (COLON ID)?)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.variable()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._maybe_50()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    def _maybe_50(self):
        """
        (COLON ID)?
        """
        pos = self.mark()
        part = self._expr_list_51()
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    def _expr_list_51(self):
        """
        (COLON ID)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expect('COLON')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('ID')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None
    @memoize
    def try_catch(self):
        pos = self.mark()
        """
        TRY try_block=compound CATCH catch_block=compound END { PyECP_Try(try_block, catch_block, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('TRY')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.compound()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('CATCH')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.compound()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('END')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            try_block = parts[1]
            catch_block = parts[3]
            return PyECP_Try(try_block, catch_block, self.loc)
        self.goto(pos)
        
        return None
        
    @memoize
    def class_definition(self):
        pos = self.mark()
        """
        CLASS name=variable body=compound END { PyECP_Class(name, body, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('CLASS')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.variable()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.compound()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expect('END')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            name = parts[1]
            body = parts[2]
            return PyECP_Class(name, body, self.loc)
        self.goto(pos)
        
        return None
        
    @memoize
    def import_statement(self):
        pos = self.mark()
        """
        IMPORT location=expr target=(AS expr)? { PyECP_Import(location, target, self.loc) };
        """
        parts = []
        for _ in range(1):
            part = self.expect('IMPORT')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self._maybe_52()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            # match:
            location = parts[1]
            target = parts[2]
            return PyECP_Import(location, target, self.loc)
        self.goto(pos)
        
        return None
        
    def _maybe_52(self):
        """
        (AS expr)?
        """
        pos = self.mark()
        part = self._expr_list_53()
        if self.match(part): return part
        self.goto(pos)
        return Filler()
    def _expr_list_53(self):
        """
        (AS expr)
        """
        pos = self.mark()
        parts = []
        for _ in range(1):
            part = self.expect('AS')
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            part = self.expr()
            if not self.match(part):
                self.fail()
                break
            parts.append(part)
            return parts
        self.goto(pos)
        return None



"""Helper functions for creating the creating python AST via the ECP parser"""

from collections import namedtuple
from parsergen import *
from parsergen.parser import ParseError
from parsergen.parser_utils import Filler, GeneratedParser
from .lexer import *
from ast import *
from typing import *
import ast
_List = ast.List
_Dict = ast.Dict

TOKEN_TO_OP = {
    "ADD":     Add,
    "SUB":     Sub,
    "MUL":     Mult,
    "DIV":     Div,
    "INT_DIV": FloorDiv,
    "POW":     Pow,
    "MOD":     Mod,
    "LT":      Lt,
    "LE":      LtE,
    "EQ":      Eq,
    "NE":      NotEq,
    "GT":      Gt,
    "GE":      GtE,
    "AND":     And,
    "OR":      Or,
    "NOT":     Not,
}

TOKEN_TO_UOP = {
    "ADD": UAdd,
    "SUB": USub,
    "NOT": Not
}

TYPE_CONVERSIONS = {
    "INT": int,
    "FLOAT": float,
    "BOOLEAN": lambda v: True if v == "True" else False,
    "STRING": str,
    "NONE": lambda v: None
}

def loc(self: GeneratedParser):
    tok = self.peek_token()
    return {"lineno": tok.start.lineno, "col_offset": tok.start.col, "end_lineno": tok.end.lineno, "end_col_offset": tok.end.col}

GeneratedParser.loc = property(loc) # add loc property so we can get location


def PyECP_Compound(c: list, l):
    if len(c) == 0:
        return [Pass(**l)]
    return c

def PyECP_Constant(t: Token, l):
    return Constant(TYPE_CONVERSIONS[t.type](t.value), **l)

def PyECP_UnaryOp(op: Token, operand):
    op_class = TOKEN_TO_UOP[op.type]()
    return UnaryOp(op=op_class, operand=operand, lineno=operand.lineno, col_offset=operand.col_offset)

def PyECP_BinOp(left, right, op: Token, l):
    return BinOp(left=left, op=TOKEN_TO_OP[op.type](), right=right, **l)

def PyECP_BoolOp(base, op, others, l):
    return BoolOp(op=TOKEN_TO_OP[op](), values=[base] + [v for _, v in others], **l)

def PyECP_Comparison(base, others, l):
    ops = [TOKEN_TO_OP[op.value]() for op, _ in others]
    comparators = [v for _, v in others]
    return Compare(left=base, ops=ops, comparators=comparators, **l)

def PyECP_ExprStatement(value):
    if isinstance(value, expr):
        return Expr(value=value, lineno=value.lineno, col_offset=value.col_offset)
    return value

def PyECP_ProcessIndexing(rv, indexes, l):
    for t, v in indexes:
        if t == "attr":
            rv = Attribute(value=rv, attr=v, ctx=Load(), **l)
        elif t == "subscript":
            rv = Subscript(value=rv, slice=Index(value=v), ctx=Load(), **l)
        elif t == "call":
            args, kwargs = v
            rv = Call(func=rv, args=args, keywords=kwargs, **l)
    return rv

def PyECP_Variable(name: Token, indexes, l):
    return PyECP_ProcessIndexing(Name(id=name.value, ctx=Load(), **l), indexes, l)

def PyECP_Assign(target, value, l):
    # TODO: allow unpacking like a,b = 1,2 (need tuple without bracket support)
    target.ctx = Store()
    return Assign(targets=[target], value=value, **l)

def PyECP_Parameters(p):
    # p: (expr (COMMA expr)* COMMA?)?
    params = []
    if not isinstance(p, Filler):
        params.append(p[0])
        for _, param in p[1]:
            params.append(param)
    return params

def PyECP_KwParameters(p):
    # p: (ID ASSIGN expr (COMMA ID ASSIGN expr)* COMMA?)?
    kw_params = []
    if not isinstance(p, Filler):
        kw_params.append(keyword(arg=p[0].value, value=p[2]))
        for _, *param in p[3]:
            kw_params.append(keyword(arg=param[0].value, value=param[2]))
    return kw_params

def PyECP_Factor(factor, indexes, l):
    rv = factor
    return PyECP_ProcessIndexing(rv, indexes, l)

def PyECP_Magic(name: str, parameters, l):
    if name == "RETURN":
        if len(parameters) == 1:
            return Return(value=parameters[0])
        if len(parameters) > 1:
            return Return(value=Tuple(elts=parameters, ctx=Load()))
        return Return()
    elif name == "CONTINUE":
        return Continue()
    elif name == "BREAK":
        return Break()
    elif name == "OUTPUT":
        name = "print"
    elif name == "USERINPUT":
        name = "input"

    return Call(func=Name(id=name, ctx=Load()), args=parameters, keywords=[], **l)


def PyECP_Tuple(values, l):
    return ast.Tuple(elts=values, ctx=Load(), **l)

def PyECP_Array(values, l):
    return ast.List(elts=values, ctx=Load(), **l)

def PyECP_Dictionary(kv_pairs, l):
    # kv_pairs: (expr COLON expr (COMMA expr COLON expr)* COMMA?)?
    keys = []
    values = []
    if not isinstance(kv_pairs, Filler):
        keys, values = [kv_pairs[0]], [kv_pairs[2]]
        for _, k, _, v in kv_pairs[3]:
            keys.append(k)
            values.append(v)
    return ast.Dict(keys=keys, values=values, **l)

def PyECP_IfStatement(condition, block, other, l):
    if isinstance(other, Filler):
        other = []
    else:
        other = other[0]
    return If(test=condition, body=block, orelse=other, **l)

def PyECP_SubroutineDef(name: Token, params, block, l):
    # params: (param_definition (COMMA param_definition)* COMMA?)?
    # TODO: keyword arguments
    parameters = []
    if not isinstance(params, Filler):
        parameters.append(params[0])
        for _, param in params[1]:
            parameters.append(param)
    return FunctionDef(
        name=name.value, 
        args=arguments(args=parameters, posonlyargs=[], kwonlyargs=[], kw_defaults=[], defaults=[], kwarg=None, vararg=None), 
        body=block, 
        decorator_list=[], 
        returns=None, 
        type_comment=None, 
        **l
    )

def PyECP_While(condition, block, l):
    return While(test=condition, body=block, orelse=[], **l)
    # TODO: implement else after while and for loops

def PyECP_RepeatUntil(condition, block, l):
    check = If(
        test=condition,
        body=[Break()],
        orelse=[],
        **l
    )

    return While(test=Constant(value=True), body=block + [check], orelse=[], **l)

def PyECP_ForTo(variable, start, end, step, block, l):
    variable.ctx = Store()
    if isinstance(step, Filler):
        step = Constant(value=1, **l)
    else:
        step = step[1]
    return For(
        target=variable,
        iter=Call(
            func=Name(id='range', ctx=Load(), **l),
            args=[
              start,
              BinOp(left=end, op=Add(), right=Constant(value=1, **l), **l),
              step
            ],
            keywords=[], **l
        ),
        body=block,
        orelse=[],
        **l
    )

def PyECP_ForIn(variable, iterator, block, l):
    variable.ctx = Store()
    return For(
        target=variable,
        iter=iterator,
        body=block,
        orelse=[],
        **l
    )

def PyECP_Record(name: Token, values, l):
    # values: (variable (COLON ID)?)*
    parameters = [v.id for v, *_ in values]
    return ClassDef(
        name=name.value,
        bases=[],
        keywords=[],
        body=[
            FunctionDef(
                name="__init__", 
                args=arguments(args=[arg(arg='self', annotation=None)]+[arg(arg=p, annotation=None) for p in parameters], posonlyargs=[], kwonlyargs=[], kw_defaults=[], defaults=[], kwarg=None, vararg=None), 
                body=[
                    Assign(
                        targets=[
                            Attribute(
                                value=Name(
                                    id='self', 
                                    ctx=Load()
                                ),
                                attr=p,
                                ctx=Store()
                            )
                        ],
                        value=Name(
                            id=p,
                            ctx=Load()
                        )
                    ) for p in parameters
                ], 
                decorator_list=[], 
                returns=None, 
                type_comment=None, 
                **l
            )
        ],
        decorator_list=[],
        **l
    )

def PyECP_Try(try_block, catch_block, l):
    return Try(
        body=try_block,
        handlers=[
            ExceptHandler(
                type=None,
                name=None,
                body=catch_block
            )
        ],
        orelse=[],
        finalbody=[],
        **l
    )
    # TODO: add support for finally and else

def PyECP_Class(name: Name, body, l):
    return ClassDef(
        name=name.id,
        bases=[],
        keywords=[],
        body=body,
        decorator_list=[],
        **l
    )

def PyECP_Import(location, target, l):
    if isinstance(target, Filler):
        target = Constant(value=None)
    else:
        target = target[1]
    
    return Expr(
        value=Call(
            func=Name(id='_ECP_IMPORT', ctx=Load()),
            args=[location, target, Call(func=Name(id='globals', ctx=Load()), args=[], keywords=[])],
            keywords=[]
        )
    )


"""ECP programming language

An implementation of the ECP programming language. Includes a Lexer and Parser for generating a python AST from
ECP code.
"""
__version__ = "1.4.5"
from .lexer import EcpLexer
from .topython import EcpParser, ecp, parse_ecp, to_py_source, _dump, get_more


from .lexer import *
from .parse import *
import argparse
import textwrap
from traceback import print_exc

class DotGenerator(NodeVisitor):
    def __init__(self, parser: Parser, shape="circle"):
        global __interpreter__
        self.parser = parser
        self.ncount = 0
        self.dot_header = [textwrap.dedent("""\
        digraph astgraph {
          node [shape="""+shape+""", fontsize=12, fontname="Courier", height=.1];
          ranksep=.3;
          edge [arrowsize=.5]
        """)]

        self.dot_body = []
        self.dot_footer = ['}']

        self.visit_IntObject = self.visit_FloatObject = self.visit_StringObject = self.visit_BoolObject = self.visit_Object

        __interpreter__ = self
    
    def addNode(self, s=None, *format_params):
        string = (s if s != None else '  node{} [label="{}"]\n').format(*format_params)
        self.dot_body.append(string)
    
    def link(self, n1, n2):
        self.dot_body.append('  node{} -> node{}\n'.format(n1, n2))
    
    def visit_Compound(self, node: Compound):
        self.addNode(None, self.ncount, "Compound")
        node._num = self.ncount
        self.ncount += 1
        
        for child in node.children:
            self.visit(child)
            self.link(node._num, child._num)
    
    def visit_Assign(self, node: Assign):
        self.visit_BinOp(node)
    
    def visit_Var(self, node: Var):
        self.addNode(None, self.ncount, node.value)
        node._num = self.ncount
        self.ncount += 1

        for p in node.array_indexes:
            self.addNode(None, self.ncount, "Property")
            self.link(node._num, self.ncount)
            property_num = self.ncount
            self.ncount += 1
            self.visit(p)
            self.link(property_num, p._num)
            
            
    
    def visit_BinOp(self, node: BinOp):
        self.addNode(None, self.ncount, node.op.value)
        node._num = self.ncount
        self.ncount += 1

        for child_node in (node.left, node.right):
            self.visit(child_node)
            self.link(node._num, child_node._num)
    
    def visit_UnaryOp(self, node: UnaryOp):
        self.addNode(None, self.ncount, "unary " + node.op.value)
        node._num = self.ncount
        self.ncount += 1

        self.visit(node.expr)
        self.link(node._num, node.expr._num)
    
    def visit_Object(self, node: Object):
        self.addNode(None, self.ncount, repr(node.value))
        node._num = self.ncount
        self.ncount += 1
    
    def visit_ArrayObject(self, node: ArrayObject):
        self.addNode(None, self.ncount, "ArrayObject")
        node._num = self.ncount
        self.ncount += 1

        for item in node.value:
            self.visit(item)
            self.link(node._num, item._num)
    
    def visit_DeclaredParam(self, node: DeclaredParam):
        self.addNode(None, self.ncount, "Param")
        node._num = self.ncount
        self.ncount += 1
        
        self.visit(node.variable)
        self.link(node._num, node.variable._num)
    
    def visit_Param(self, node: Param):
        self.addNode(None, self.ncount, "Param")
        node._num = self.ncount
        self.ncount += 1

        self.visit(node.value)
        self.link(node._num, node.value._num)


    def visit_Subroutine(self, node: Subroutine):
        self.addNode(
            '  node{} [label="Subroutine:{}"]\n',
            self.ncount,
            node.token.value
        )
        node._num = self.ncount
        self.ncount += 1

        for param_node in node.parameters:
            self.visit(param_node)
            self.link(node._num, param_node._num)
        
        self.visit(node.compound)
        self.link(node._num, node.compound._num)
    
    def visit_SubroutineCall(self, node: SubroutineCall):
        self.addNode(None, self.ncount, "SubroutineCall")
        node._num = self.ncount
        self.ncount += 1

        self.addNode(None, self.ncount, "Target")
        self.link(node._num, self.ncount)
        t_num = self.ncount
        self.ncount += 1

        self.visit(node.subroutine_token)
        self.link(t_num, node.subroutine_token._num)


        for param_node in node.parameters:
            self.visit(param_node)
            self.link(node._num, param_node._num)
    
    def visit_Record(self, node: Record):
        self.addNode('  node{} [label="Record:{}"]\n', self.ncount, node.token.value)
        node._num = self.ncount
        self.ncount += 1
        
        for p in node.parameters:
            self.addNode(None, self.ncount, p)
            _num = self.ncount
            self.ncount += 1
            
            self.link(node._num, _num)
    
    def visit_ClassDefinition(self, node: ClassDefinition):
        self.addNode('  node{} [label="ClassDefinition:{}"]\n', self.ncount, node.token.value)
        node._num = self.ncount
        self.ncount += 1

        for v in node.static_values:
            self.visit(v)
            self.link(node._num, v._num)
        
        for subroutine in node.subroutines:
            self.visit(subroutine)
            self.link(node._num, subroutine._num)
    
    def visit_Magic(self, node: Magic):
        self.addNode('  node{} [label="Magic:{}"]\n', self.ncount, node.token.value)
        node._num = self.ncount
        self.ncount += 1

        for param_node in node.parameters:
            self.visit(param_node)
            self.link(node._num, param_node._num)
    
    def visit_IfStatement(self, node: IfStatement):
        self.addNode(None, self.ncount, "IfStatement")
        node._num = self.ncount
        self.ncount += 1

        self.addNode(None, self.ncount, "Condition")
        self.link(node._num, self.ncount)
        condition_num = self.ncount
        self.ncount += 1

        self.visit(node.condition)
        self.link(condition_num, node.condition._num)

        self.addNode(None, self.ncount, "Consequence")
        self.link(node._num, self.ncount)
        consequence_num = self.ncount
        self.ncount += 1

        self.visit(node.consequence)
        self.link(consequence_num, node.consequence._num)

        self.addNode(None, self.ncount, "Alternatives")
        self.link(node._num, self.ncount)
        alternatives_num = self.ncount
        self.ncount += 1
        
        for statement in node.alternatives:
            self.visit(statement)
            self.link(alternatives_num, statement._num)
    
    def visit_WhileStatement(self, node: WhileStatement, isRepeatUntilStatement=False):
        name = "RepeatUntil" if isRepeatUntilStatement else "While"
        self.addNode(None, self.ncount, name+"Statement")
        node._num = self.ncount
        self.ncount += 1

        self.addNode(None, self.ncount, "Condition")
        self.link(node._num, self.ncount)
        condition_num = self.ncount
        self.ncount += 1

        self.visit(node.condition)
        self.link(condition_num, node.condition._num)

        self.addNode(None, self.ncount, "Consequence")
        self.link(node._num, self.ncount)
        consequence_num = self.ncount
        self.ncount += 1

        self.visit(node.consequence)
        self.link(consequence_num, node.consequence._num)
    
    def visit_RepeatUntilStatement(self, node: RepeatUntilStatement):
        self.visit_WhileStatement(node, True)
    
    def visit_ForLoop(self, node: ForLoop):
        self.addNode(None, self.ncount, "ForLoop")
        node._num = self.ncount
        self.ncount += 1

        self.addNode(None, self.ncount, "Variable")
        self.link(node._num, self.ncount)
        var_num = self.ncount
        self.ncount += 1

        self.visit(node.variable)
        self.link(var_num, node.variable._num)

        if node.to_step:
            self.addNode(None, self.ncount, "Start")
            self.link(node._num, self.ncount)
            start_num = self.ncount
            self.ncount += 1

            self.visit(node.start)

            self.link(start_num, node.start._num)

            self.addNode(None, self.ncount, "End")
            self.link(node._num, self.ncount)
            end_num = self.ncount
            self.ncount += 1

            self.visit(node.end)

            self.link(end_num, node.end._num)

            if node.step:
                self.addNode(None, self.ncount, "Step")
                self.link(node._num, self.ncount)
                step_num = self.ncount
                self.ncount += 1

                self.visit(node.step)

                self.link(step_num, node.step._num)
        else:
            self.addNode(None, self.ncount, "Iterator")
            self.link(node._num, self.ncount)
            iterator_num = self.ncount
            self.ncount += 1

            self.visit(node.iterator)
            self.link(iterator_num, node.iterator._num)

        
        self.addNode(None, self.ncount, "Loop")
        self.link(node._num, self.ncount)
        loop_num = self.ncount
        self.ncount += 1
        self.visit(node.loop)
        self.link(loop_num, node.loop._num)
    
    def visit_TryCatch(self, node: TryCatch):
        self.addNode(None, self.ncount, "TryCatch")
        node._num = self.ncount
        self.ncount += 1

        self.addNode(None, self.ncount, "Try")
        _num = self.ncount
        self.link(node._num, _num)
        self.ncount += 1
        self.visit(node.try_compound)
        self.link(_num, node.try_compound._num)
    
        self.addNode(None, self.ncount, "Catch")
        _num = self.ncount
        self.link(node._num, _num)
        self.ncount += 1
        self.visit(node.catch_compound)
        self.link(_num, node.catch_compound._num)
    
    def visit_NoOp(self, node: NoOp):
        self.addNode(None, self.ncount, "NoOp")
        node._num = self.ncount
        self.ncount += 1
    
    def generate(self):
        tree = self.parser.parse()
        self.visit(tree)

        return ''.join(self.dot_header + self.dot_body + self.dot_footer)



parser = argparse.ArgumentParser(description="Dot file generator")
parser.add_argument("inputfile", type=argparse.FileType("r", encoding="utf-8"), nargs="?")
parser.add_argument("-c", nargs="?")
parser.add_argument("--shape", action="store", dest="shape")

options = parser.parse_args()
#print(options)
if options.inputfile:
    s = options.inputfile.read()
    options.inputfile.close()
else:
    s = options.c

l = Lexer()
result = l.lexString(s)
p = Parser(result)
i = DotGenerator(p, shape=options.shape if options.shape else "circle")
s = i.generate()
print(s)

# Example:
# python genasdot.py -c "OUTPUT 2 + 3 * 4" > ast.dot && dot -Tpng -o ast.png ast.dot
# python genasdot.py ..\examples\sieve.ecp > ast.dot && dot -Tpng -o ast.png ast.dot
    


from tabulate import tabulate
from typing import *
from collections.abc import Callable
import sys

class VarContainer:
    def __init__(self, onchange: Callable):
        self._onchange = onchange

    def __setattr__(self, name, value):
        if name != "_onchange":
            self._onchange(name, value)
        super(VarContainer, self).__setattr__(name, value)

class Tracker:
    def __init__(self, variables: List[str], compact=False):
        self.values = {}
        self.line = 0
        self.compact = compact
        self.variables = variables
    
    def onchange(self, name, value):
        #print(f"{name} -> {value}")
        if name in self.variables:
            data: dict = self.values.get(name, {})
            
            #while self.line in data.keys():
            if data.get(self.line) != None or not self.compact:
                self.line += 1
            
            data[self.line] = value
    
            self.values[name] = data
    def displayTraceTable(self, variables: List[str] = None, tablefmt="github") -> str:
        headers = [k for k in self.variables if k in self.values]
        if variables:
            headers = variables
        
        maxLine = 0
        for v in self.values.values():
            for line in v.keys():
                if line > maxLine:
                    maxLine = line
        
        
        table = []
        for i in range(maxLine+1):
            row = []
            for h in headers:
                row.append(self.values.get(h, {}).get(i, ""))
            table.append(row)
        return tabulate(table, headers=headers, tablefmt=tablefmt)

class Tracer(Tracker):
    def __init__(self, variables: List[str], **kwargs):
        super().__init__(variables, **kwargs)
        self.variables = variables
    
    def tracer(self, frame, event, arg = None):
        code = frame.f_code
        line_no = frame.f_lineno
        #self.onchange("_line_number", line_no)
        for v in self.variables:
            if len(self.values.get(v, {}).keys()) > 0:
                if self.values[v][max(self.values[v].keys())] == frame.f_locals.get(v):
                    continue
            if v in frame.f_locals.keys():
                self.onchange(v, frame.f_locals.get(v))
        return self.tracer

    def __enter__(self):
        sys.settrace(self.tracer)
        return self
    
    def __exit__(self, *args):
        print(self.displayTraceTable())
        sys.settrace(None)


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("src/ecp/__init__.py") as f:
    lines = f.readlines()

version = None
for l in lines:
    if l.startswith("__version__"):
        version = l.split("=")[1].strip()[1:-1]

if version is None:
    raise Exception("Version not found!")

setuptools.setup(
    name="py-ecp",
    version=version,
    author="Conqu3red",
    description="ECP programming language",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ecp-lang/ecp-interpreter",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=[
        "tabulate",
        "astor",
        "parsergen==2.0.0b8"
    ],
    entry_points = {
        'console_scripts': [
            'ecp = ecp.__main__:main'
        ]
    },
)


from sys import exc_info
from .lexer import *
from .topython import *
from . import __version__, _dump
import argparse
from .tracker import Tracker
from traceback import print_exc
import os

def main():
    parser = argparse.ArgumentParser("ecp", description="ECP interpreter")
    parser.add_argument("inputfile", type=argparse.FileType("r", encoding="utf-8"), nargs="?")
    parser.add_argument("--debug", action="store_true", help="show debug information like token list")
    parser.add_argument("--showast", action="store_true", help="show AST")
    parser.add_argument("--trace", action="store", nargs="*", default=[], help="space seperated names of the variables to be traced")
    parser.add_argument("--tracecompact", action="store_true", help="trace compactly")
    parser.add_argument("--topython", action="store_true", help="Try to convert the ECP program to python source code")
    parser.add_argument("--pause", action="store_true", help="pause on completion")
    parser.add_argument('--version', action='version', version='%(prog)s v'+__version__)

    options = parser.parse_args()
    should_trace = len(options.trace) > 0
    #print(options)
    if options.inputfile:
        string = options.inputfile.read()
        loc = os.path.dirname(os.path.abspath(options.inputfile.name))
        name = os.path.basename(options.inputfile.name)
        options.inputfile.close()

        def debugOutput(result):
            table = []
            for i in result.tokens:
                table.append([i.value, i.type])
            print(tabulate(table, tablefmt="github", headers=["VALUE", "TYPE"]))

        if options.debug:
            pass
            #debugOutput(result)
        sys.path.insert(0, loc)
        if options.topython:
            print(to_py_source(string))
        else:
            #print(_dump(parse_ecp(string), indent=2, include_attributes=True)) # DEBUG
            ecp(string, name=name, scope=globals(), trace=options.trace, tracecompact=options.tracecompact, showAST=options.showast)
        if options.pause:
            input("Press enter to exit...")

    else:
        # Live console
        print(f"ECP {__version__}")
        string = ""
        prompt = "ECP> "
        more_prompt = ".... "
        while True:
            try:
                string = input(prompt)
                more = get_more(string)
                while more:
                    string += "\n" + input(more_prompt)
                    #print(repr(string))
                    more = get_more(string)
                try:
                    ecp(string, name="<stdin>", scope=globals(), mode="single")
                except SystemExit:
                    exit()
                except:
                    print_exc(0)
            except KeyboardInterrupt:
                print()
                string = ""
            except EOFError:
                exit()

if __name__ == "__main__":
    main()


from typing import *
from parsergen.lexer import *
from tabulate import tabulate
import codecs

def use_name(name, *rules):
    """Helper function for crating tokens which return their type as value
    
    Usage:
    >>> EQ = use_name("EQ", r"=")
    """
    def modifier(self, t):
        t.value = name
        return t
    
    modifier.__name__ = name

    return token(*rules)(modifier)

class EcpLexer(Lexer):
    # symbols
    ASSIGN  = r"←", r":="
    EQ      = use_name("EQ",      r"="           )
    ADD     = use_name("ADD",     r"\+"          )
    POW     = use_name("POW",     r"\*\*", r"POW")
    MUL     = use_name("MUL",     r"\*"          )
    SUB     = use_name("SUB",     r"\-",   r"–"  )
    INT_DIV = use_name("INT_DIV", r"DIV"         )
    MOD     = use_name("MOD",     r"%",    r"MOD")
    DIV     = use_name("DIV",     r"/"           )
    NE      = use_name("NE",      r"!=",   r"≠"  )
    LE      = use_name("LE",      r"<=",   r"≤"  )
    GE      = use_name("GE",      r">=",   r"≥"  )
    LT      = use_name("LT",      r"<"           )
    GT      = use_name("GT",      r">"           )
    NOT     = use_name("NOT",     r"NOT"         )
    OR      = use_name("OR",      r"OR"          )
    AND     = use_name("AND",     r"AND"         )
    
    @token(r"\n")
    def NEWLINE(self, t):
        self.lineno += len(t.value)
        self.column = 0
        return None
    
    LPAREN = r"\("
    RPAREN = r"\)"
    LS_PAREN = r"\["
    RS_PAREN = r"\]"
    LC_BRACE = r"\{"
    RC_BRACE = r"\}"
    COMMA = r","
    COLON = r"\:"
    
    # types
    @token(r"(\d*\.\d+)", r"(\d+\.\d*)")
    def FLOAT(self, t):
        float(t.value)
        return t
    
    @token(r"\d+")
    def INT(self, t):
        int(t.value)
        return t
    
    BOOLEAN = r"True", r"False"

    @token(r"(\"|')")
    def STRING(self, t):
        end = t.value
        escape = False
        result = ""
        while True:
            if len(self.source) > 0:
                cur_char = self.source[0]
                self.step_source(1)
                if not escape and cur_char == "\\":
                    result += cur_char
                    if len(self.source) == 0:
                        continue
                    cur_char = self.source[0]
                    self.step_source(1)
                    escape = True
                if cur_char == end and not escape:
                    t.value = codecs.getdecoder("unicode_escape")(result)[0]
                    t.end = Pos(self.lineno, self.column)
                    return t
                result += cur_char
                escape = False
            else:
                raise Exception("No end of string")
    
    
    NONE = r"None"

    DOT = r"\."
    
    # keywords
    SUBROUTINE = r"SUBROUTINE",
    END = r"ENDSUBROUTINE", r"ENDIF", r"ENDWHILE", r"ENDFOR", r"ENDRECORD", r"ENDTRY", r"ENDCLASS", r"END"
    MAGIC = r"RETURN", r"CONTINUE", r"BREAK", r"OUTPUT", r"USERINPUT"
    IF = r"IF"
    THEN = r"THEN"
    ELSE = r"ELSE"
    WHILE = r"WHILE"
    REPEAT = r"REPEAT"
    UNTIL = r"UNTIL"
    FOR = r"FOR"
    TO = r"TO"
    IN = r"IN"
    STEP = r"STEP"
    RECORD = r"RECORD"
    CONSTANT = r"CONSTANT"
    TRY = r"TRY"
    CATCH = r"CATCH"
    CLASS = r"CLASS"
    IMPORT = r"IMPORT"
    AS = r"AS"

    ID = r"[a-zA-Z_][a-zA-Z0-9_]*"

    ignore = " \t"
    ignore_comment = r"#.*"